from DTUtils import Logger
from pathlib import Path
import glob
import requests
from datetime import datetime
from ruxit.api.base_plugin import BasePlugin
from ruxit.api.exceptions import ConfigException
import time
import urllib3
import socket
import os

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


attempts_max = 3
sleep_time = 5

# cannot use as doesn't work when dt.entity.host gets added. The problem becomes associated to a host and the entity dissapears.
ingestURL = "http://localhost:14499/metrics/ingest"


filePath = '''[
  {
    "schemaId": "builtin:monitoredentities.generic.type",
    "scope": "environment",
    "value": {
        "enabled": true,
        "name": "filechecker:path",
        "displayName": "File system path",
        "createdBy": "extension",
        "rules": [
          {
            "idPattern": "{path}",
            "instanceNamePattern": "{path}",
            "sources": [
              {
                "sourceType": "Metrics",
                "condition": "$prefix(filechecker.)"
              }
            ],
            "requiredDimensions": [],
            "attributes": []
          }
        ]
    }
  }
]'''

# not used because of issue described above. Will re-instanciate when fixed.
fileRelationship ='''[
    {
    "schemaId": "builtin:monitoredentities.generic.relation",
    "scope": "environment",      
      "value": {
        "enabled": true,
        "sources": [
          {
            "sourceType": "Metrics",
            "condition": "$prefix(filechecker.)"
          }
        ],
        "createdBy": "extension",
        "fromType": "filechecker:path",
        "typeOfRelation": "CHILD_OF",
        "toType": "host"
      }
    }
]
'''

class fileCheckerMINT(BasePlugin):
    def initialize(self, **kwargs):
        self.logger = Logger(bool(self.config['debug']), __name__)

        self.executions = 0

        self.paths_files = [path.strip() for path in self.config.get("paths_files").split(",") if path]
        self.checkWildcard(self.paths_files, "File exists")
        self.paths_folders = [path.strip() for path in self.config.get("paths_folders").split(",") if path]
        self.checkWildcard(self.paths_folders, "Folder exists")


        # path age can operate as path1,path2 or path1|1m,path2 where path1 won't return the file age in second but the number of files which are older than 1 minute. 
        # Uses a different metric file_age_count instead of file_age
        self.paths_age = self.splitList([path.strip() for path in self.config.get("paths_age").split(",") if path])
        self.paths_size = [path.strip() for path in self.config.get("paths_size").split(",") if path]
        self.paths_count = [path.strip() for path in self.config.get("paths_count").split(",") if path]

        self.paths_mount = [path.strip() for path in self.config.get("paths_mount").split(",") if path]

        self.latest_file = self.config.get("latest_file")

        self.tenant_url = self.config.get("tenant_url")
        if self.tenant_url.strip() == "":
            raise ConfigException("Tenant URL cannot be empty")
        if self.tenant_url[-1] != "/":
            self.tenant_url += "/"

        self.include_folders = self.config.get("include_folders")
        self.debug = self.config.get("debug", False)

        self.tenant_token = self.config.get("tenant_token")
        if self.tenant_token.strip() == "":
            raise ConfigException("Token cannot be empty")

        self.createEntities()    
        self.hostname = socket.gethostname()    


    def splitList(self, myList):
        tmpList = []
        for item in myList:
            # check for optional separator
            if item.find("|") != -1:
                tmpValue = item.split("|")
                tmpList.append([tmpValue[0], self.convertToSeconds(tmpValue[1])])
            else:
                tmpList.append([item,0])
        return tmpList

    # convert a string such as 10s to the int value 10.
    # accepts s,m,h,d (seconds, minutes, hours, days) The value is converted to seconds.
    def convertToSeconds(self, duration):
        if len(duration) < 2:
            raise ConfigException("Incorrect value in file age, must be at least 2 characters long such as 1s, 2d etc...")
        unit = duration[-1]
        if unit not in ['s','m','h','d']:
            raise ConfigException("Incorrect unit in file age field, must be s,m,h or d: " + str(duration))
        value = duration[0:-1]
        try:
            intValue = int(value)
            if unit == "s":
                return intValue
            elif unit == "m":
                return intValue*60
            elif unit == "h":
                return intValue*60*60
            elif unit == "d":
                return intValue*60*60*24
            

        except Exception as e:
            raise ConfigException("Incorrect value in file age. The duration must be a number: " + duration)




    # create if necessary the generic data types filechecker:path
    def createEntities(self):
            try:
                foundPath = False
                response = requests.get(self.tenant_url + "api/v2/settings/objects?schemaIds=builtin%3Amonitoredentities.generic.type&scopes=environment&fields=objectId%2Cvalue", headers={"Content-Type": "application/json; charset=utf-8", "Authorization": "Api-Token " + self.tenant_token}, timeout=5, verify=False)
                if response.status_code == 404:
                    self.logger.info("No entity types present " + str(response.status_code) + ":" + response.text )
                elif response.status_code == 429:
                    self.logger.info("HTTP 429 when querying entities.")
                else:
                    output = response.json()
                    for entry in output["items"]:
                        if entry["value"]["name"] == "filechecker:path" and entry["value"]["createdBy"] == "extension":
                            foundPath = True

                    # assumes that path and relationship have both been successful. This is to reduce the number of unncessary API calls.
                    if foundPath is False:
                        self.createSettingsObject(filePath)
                        #self.createSettingsObject(fileRelationship)
                    else:
                        self.logger.info("filechecker:path already present")

            except Exception as ex:
                self.logger.info("Cannot send data to query settings objects " + str(ex))  
                self.logger.exception(ex)   


    def createSettingsObject(self,myData):
        try:
            response = requests.post(self.tenant_url + "api/v2/settings/objects", data=myData, headers={"Content-Type": "application/json; charset=utf-8", "Authorization": "Api-Token " + self.tenant_token}, timeout=5, verify=False)
            if response.status_code == 200:
                self.logger.info("Setting object created" )
            else:
                self.logger.error("Setting object couldn't be created " + str(response.status_code))
            
        except Exception as ex:
            self.logger.info("Cannot create object settings " + str(ex))  
            self.logger.exception (ex)           

    # some fields don't accept wild card. Block them to avoid confusions.
    def checkWildcard(self, data, fieldName):
        for item in data:
            if self.isGlob(item) is True:
                raise ConfigException("Wildcard pattern detected. '*, [, ], ?, !' are not allowed. " + fieldName + " only accepts exact paths")


    # Checks if one of the reserved characters for glob are in the string.
    def isGlob(self, data):
        if data.find("*") == -1 and data.find("[") and data.find("?") == -1 and data.find("!") == -1:
            return False
        else:
            return True

    # Takes a path and checks that there is a pattern match (glob format). If not add, a standard wildcard.
    # Should only be used for file count, file size and file age which typically return multiple files
    # This is to cover scenarios such as /tmp/ -> return all the files in /tmp/ so replaced by /tmp/*
    # Glob will accept * ? [] and ! for negating ranges. Glob is NOT regEx.
    def globify(self, data):
        if self.isGlob(data) is False:    
            return data + "*"
        else:
            return data



    # returns the last modified date. If more than 1 file returns the oldest file matching the path if latest_file is False.
    # returns -1 if nothing matches
    def last_modified(self, path: str) -> int:
        lastModified = -1
        # check if contains a wildcard. If yes use the glob function otherwise only check that the file exists
        if self.isGlob(path) is True :
            myPath = self.globify(path)        
            for f in glob.glob(myPath, recursive=True):
                try:
                    if Path(f).is_file():
                        tmp = int((datetime.now() - datetime.fromtimestamp(Path(f).stat().st_mtime)).total_seconds())
                        if self.latest_file is False:
                            if tmp > lastModified:
                                lastModified = tmp
                        else:
                            if lastModified == -1:
                                lastModified = tmp
                            elif tmp < lastModified:
                                lastModified = tmp

                except Exception as e:
                    self.logger.error("Could not get last modified for {f}: {e}")
                    self.logger.exception(e)
            return lastModified
        else:
            if Path(path).exists() and Path(path).is_file():
                return int((datetime.now() - datetime.fromtimestamp(Path(path).stat().st_mtime)).total_seconds())
            else:
                return -1


    # Returns the number of files which are older than the duration
    # returns 0 if nothing matches
    def last_modified_count(self, path: str, duration: int) -> int:
        fileCount = 0
        # check if contains a wildcard. If yes use the glob function otherwise only check if the file is older than the duration
        if self.isGlob(path) is True :
            myPath = self.globify(path)        
            for f in glob.glob(myPath, recursive=True):
                try:
                    if Path(f).is_file():
                        tmp = int((datetime.now() - datetime.fromtimestamp(Path(f).stat().st_mtime)).total_seconds())
                        if tmp > duration:
                            fileCount += 1
                except Exception as e:
                    self.logger.error("Could not get last modified for {f}: {e}")
                    self.logger.exception(e)
            return fileCount
        else:
            if Path(path).exists() and Path(path).is_file():
                tmp = int((datetime.now() - datetime.fromtimestamp(Path(path).stat().st_mtime)).total_seconds())
                if tmp > duration:
                    return 1
                else:
                    return 0
            else:
                return 0

    def size(self, path: str) -> int:
        sum_sizes = 0
        myPath = self.globify(path)

        for f in glob.glob(myPath, recursive=True):
            try:
                if Path(f).is_file():
                    sum_sizes += Path(f).stat().st_size
            except Exception as e:
                self.logger.error(f"Could not get size for {f}: {e}")
                self.logger.exception(e)
        return sum_sizes

    def count(self, path: str) -> int:

        myPath = self.globify(path)

        if self.include_folders is True:
            return len([f for f in glob.glob(myPath, recursive=True)])
        else:
            return len([f for f in glob.glob(myPath, recursive=True) if Path(f).is_file()])


    # Unix only, checks that the mount point is valid.
    def mounted(self, path: str) -> int:
        if Path(path).is_dir() and os.path.ismount(path):
            return 1
        else:
            return 0



    # send all the metrics to mint
    def sendDataToMint(self, url , postData):

        if len(postData)> 0:
            try:
                attempts = 1
                while attempts <= attempts_max:
                    response = requests.post(ingestURL, data=postData, headers={"Content-Type": "text/plain; charset=utf-8", "eec-enrichment": "false"}, timeout=5, verify=False)
                    # response = requests.post(url, data=postData, headers={"Content-Type": "text/plain; charset=utf-8", "Authorization": "Api-Token " + self.tenant_token}, timeout=5, verify=False)
                    
                    if response.status_code == 202:
                        self.logger.debug("Mint post successful")
                        break
                    elif response.status_code == 429:    
                        self.logger.info("Too many requests error from mint, sleeping")
                        attempts += 1
                        time.sleep(sleep_time)

                    else:
                        self.logger.info("Data couldn't be sent to MINT" + str(response.status_code) + ":" + response.text )
                        break
                if attempts > attempts_max:
                    self.logger.info("MINT data not sent due to 429")

            except Exception as ex:
                self.logger.exception(ex)
        else:
            self.logger.debug("No data to send to mint")


    def query(self, **kwargs) -> None:

        mintMetrics = ""

        frequency = self.config.get("frequency", 1)
        if self.executions % frequency == 0:
            self.logger.info(f"Running the file checks")
                
            for path in self.paths_files:
                if not Path(path).exists():
                    mintMetrics += "filechecker.file_exists,path=\"" + path + "\",hostname=\"" + self.hostname + "\" 0\n"
                else:
                    mintMetrics += "filechecker.file_exists,path=\"" + path + "\",hostname=\"" + self.hostname + "\" 1\n"

            for path in self.paths_folders:
                if not Path(path).exists():
                    mintMetrics += "filechecker.folder_exists,path=\"" + path + "\",hostname=\"" + self.hostname + "\" 0\n"
                else:
                    mintMetrics += "filechecker.folder_exists,path=\"" + path + "\",hostname=\"" + self.hostname + "\" 1\n"
        
            for path in self.paths_size:
                mintMetrics += "filechecker.file_size,path=\"" + path + "\",hostname=\"" + self.hostname + "\" " + str(self.size(path)) + "\n"

            for path, duration in self.paths_age:
                if duration == 0:
                    mintMetrics += "filechecker.file_age,path=\"" + path + "\",hostname=\"" + self.hostname + "\" " + str(self.last_modified(path)) + "\n"
                else:
                    self.logger.debug(path + ":" + str(duration))
                    mintMetrics += "filechecker.file_age_count,path=\"" + path + "\",hostname=\"" + self.hostname + "\" " + str(self.last_modified_count(path, duration)) + "\n"

            for path in self.paths_count:
                mintMetrics += "filechecker.file_count,path=\"" + path + "\",hostname=\"" + self.hostname + "\" " + str(self.count(path)) + "\n"

            for path in self.paths_mount:
                mintMetrics += "filechecker.file_mount,path=\"" + path + "\",hostname=\"" + self.hostname + "\" " + str(self.mounted(path)) + "\n"

            if len(mintMetrics)>0:
               
                mintMetrics = mintMetrics.replace("\\", "\\\\")
                metaData = "#filechecker.file_exists gauge dt.meta.unit=count\n"
                metaData += "#filechecker.folder_exists gauge dt.meta.unit=count\n"
                metaData += "#filechecker.file_age gauge dt.meta.unit=seconds\n"
                metaData += "#filechecker.file_age_count gauge dt.meta.unit=count\n"                
                metaData += "#filechecker.file_size gauge dt.meta.unit=bytes\n"
                metaData += "#filechecker.file_count gauge dt.meta.unit=count\n"
                mintMetrics += metaData
                self.logger.debug(mintMetrics)
                self.sendDataToMint(self.tenant_url + "api/v2/metrics/ingest", mintMetrics)

        else:
            self.logger.info(f"Not running. {frequency - (self.executions % frequency)} minutes remaining for next run")

        self.executions += 1
